# -*- coding: utf-8 -*-
"""Management command to create a new AccessToken."""
import datetime
from argparse import ArgumentParser
from typing import Any

from django.core.management.base import BaseCommand
from django.db.utils import IntegrityError
from django.utils.timezone import now

from perimeter.models import AccessToken
from perimeter.settings import PERIMETER_DEFAULT_EXPIRY


class Command(BaseCommand):

    help = "Create a perimeter access token."

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument(
            "-t",
            "--token",
            action="store",
            dest="token",
            help="User supplied token (max 10 chars)",
        )
        parser.add_argument(
            "-e",
            "--expires",
            type=int,
            action="store",
            dest="expires",
            help="Expires value (in days)",
        )

    def handle(self, *args: Any, **options: Any) -> None:

        has_expires = options.get("expires") is not None
        days = options.get("expires") or PERIMETER_DEFAULT_EXPIRY
        token = options.get("token") or AccessToken.random_token_value()
        expires_on = (now() + datetime.timedelta(days=days)).date()
        try:
            access_token = AccessToken.objects.create_access_token(
                token=token, expires_on=expires_on
            )
            self.stdout.write(
                'Created new access token: "{}" (expires {})'.format(
                    access_token.token, access_token.expires_on
                )
            )
        except IntegrityError:
            access_token = AccessToken.objects.get(token=token)
            if has_expires:
                self.stdout.write("Extending existing token")
                access_token.expires_on = expires_on
                access_token.save()
                self.stdout.write("Token extended: {}".format(access_token))
            else:
                self.stdout.write(
                    "Token exists already - please use --expires option to extend."
                )



"""
Check all incoming requests for a valid token.

See Perimeter docs for more details.

"""
from typing import Any, Callable, Optional, Union
from urllib.parse import urlencode

from django.core.exceptions import ImproperlyConfigured, MiddlewareNotUsed
from django.http import HttpRequest, HttpResponse, HttpResponseRedirect
from django.urls import reverse
from django.utils.deprecation import MiddlewareMixin

from .models import AccessToken, EmptyToken
from .settings import HTTP_X_PERIMETER_TOKEN
from .settings import PERIMETER_BYPASS_FUNCTION as bypass_perimeter
from .settings import PERIMETER_ENABLED, PERIMETER_SESSION_KEY


def check_middleware(func: Callable) -> Callable:
    """Check a request arg has a Session attached."""
    # noqa: blank line to make black and flake8 play nicely
    def inner(request: HttpRequest, *args: Any) -> Optional[HttpResponse]:
        if not hasattr(request, "session"):
            raise ImproperlyConfigured(
                "Missing session attribute - please check MIDDLEWARE_CLASSES for "
                "'django.contrib.sessions.middleware.SessionMiddleware'."
            )
        return func(request, *args)

    return inner


@check_middleware
def get_request_token(request: HttpRequest) -> Optional[str]:
    """Extract token string from HTTP header or querystring."""
    return request.META.get(HTTP_X_PERIMETER_TOKEN, None) or request.session.get(
        PERIMETER_SESSION_KEY, None
    )


@check_middleware
def set_request_token(request: HttpRequest, token_value: str) -> None:
    """Set the request.session token value."""
    request.session[PERIMETER_SESSION_KEY] = token_value


def get_access_token(request: HttpRequest) -> Union[AccessToken, EmptyToken]:
    """Fetch the AccessToken from the request."""
    token_value = get_request_token(request)
    return AccessToken.objects.get_access_token(token_value)


def get_redirect_url(request: HttpRequest) -> str:
    """Unpack request and extract a valid redirect url."""
    qstring = urlencode({"next": request.get_full_path()})
    url = reverse("perimeter:gateway")
    return f"{url}?{qstring}"


class PerimeterAccessMiddleware(MiddlewareMixin):
    """
    Middleware used to detect whether user can access site or not.

    This middleware will be disabled if the PERIMETER_ENABLED setting does not
    exist in django settings, or is False.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """
        Disable middleware if PERIMETER_ENABLED setting not True.

        Raises MiddlewareNotUsed exception if the PERIMETER_ENABLED setting
        is not True - this is used by Django framework to remove the middleware.
        """
        if PERIMETER_ENABLED is False:
            raise MiddlewareNotUsed("Perimeter disabled")
        super().__init__(*args, **kwargs)

    def process_request(self, request: HttpRequest) -> Optional[HttpResponseRedirect]:
        """Check user session for token."""
        if bypass_perimeter(request):
            return None

        if get_access_token(request).is_valid:
            return None

        return HttpResponseRedirect(get_redirect_url(request))



# -*- coding: utf-8 -*-
"""Management command to list all active tokens."""
from typing import Any

from django.core.management.base import BaseCommand

from perimeter.models import AccessToken


class Command(BaseCommand):

    help = "List all active tokens."

    def handle(self, *args: Any, **options: Any) -> None:

        self.stdout.write("Listing all tokens:")
        for token in AccessToken.objects.all():
            prefix = "- " if token.is_valid else "x "
            suffix = " expired " if token.has_expired else " expires "
            self.stdout.write(f"{prefix} {token} {suffix} {token.expires_on}")



from urllib.parse import unquote

from django.http import HttpResponseRedirect
from django.http.request import HttpRequest
from django.http.response import HttpResponse, HttpResponseNotAllowed
from django.shortcuts import render
from django.urls import Resolver404, resolve, reverse

from .forms import TokenGatewayForm, UserGatewayForm
from .settings import PERIMETER_REQUIRE_USER_DETAILS


def resolve_return_url(return_url: str) -> str:
    """
    Resolve a URL to confirm that it's valid.

    Before redirecting to a return_url, use the resolve function
    to confirm that it matches a valid view function.

    If the resolver can't find the function, return the only url we
    know exists - perimeter:gateway

    """
    path = None
    if return_url:
        return_url = unquote(return_url)
        # Path with a query string will not resolve, so we strip it for the check.
        path = return_url.split("?")[0]

    try:
        resolve(path)
        return return_url
    except Resolver404:
        return reverse("perimeter:gateway")


def gateway(
    request: HttpRequest, template_name: str = "perimeter/gateway.html"
) -> HttpResponse:
    """
    Display gateway form and process access requests.

    When the PerimeterAccessMiddleware catches an unvalidated
    user request they will redirect to this page.

    """
    # the form to use is based on whether we want user details or not.
    klass = UserGatewayForm if PERIMETER_REQUIRE_USER_DETAILS else TokenGatewayForm

    if request.method == "GET":
        form = klass()

    elif request.method == "POST":
        form = klass(request.POST)
        if form.is_valid():
            form.save(request)
            return HttpResponseRedirect(resolve_return_url(request.GET.get("next")))

    else:
        return HttpResponseNotAllowed(["GET", "POST"])

    return render(request, template_name, {"form": form})



from os import environ

from django.conf import settings
from django.urls import reverse

CAST_AS_BOOL = lambda x: x in (True, "true", "True")  # noqa: E731
CAST_AS_INT = lambda x: int(x)  # noqa: E731


def get_setting(setting_name, default_value, cast_func=lambda x: x):
    """Return setting from django.conf or os.environ.

    This function will look up a setting in the os.environ first, and
    then if not found there, the django.conf.settings. Values from env
    vars are strings by default, so the cast_func can be used to convert
    strings into booleans or integers etc.

    Args:
        setting_name: string - the name of the setting to fetch
        cast_func: function - a function that is applied to the value,
            commonly used to cast string values to a bool, int etc.
        default_value: value to use if the setting_name is not found

    """
    return cast_func(
        environ.get(setting_name) or getattr(settings, setting_name, default_value)
    )


# Name of HTTP header used to automatically bypass perimeter
HTTP_X_PERIMETER_TOKEN = "HTTP_X_PERIMETER_TOKEN"  # noqa: S105

# if False, the middleware will be disabled
PERIMETER_ENABLED = get_setting("PERIMETER_ENABLED", False, cast_func=CAST_AS_BOOL)
# request.session key used to store user's token
PERIMETER_SESSION_KEY = get_setting("PERIMETER_SESSION_KEY", "perimeter")
# default expiry, in days, of a token
PERIMETER_DEFAULT_EXPIRY = get_setting(
    "PERIMETER_DEFAULT_EXPIRY", 7, cast_func=CAST_AS_INT
)
# function used to bypass the perimter - must be function that takes request
# as only arg.
# NB we don't use get_setting here - you can't put a function into an env var
PERIMETER_BYPASS_FUNCTION = getattr(
    settings,
    "PERIMETER_BYPASS_FUNCTION",
    # default function is to restrict everything except the gateway page itself
    lambda r: r.path == reverse("perimeter:gateway"),
)
# If True, then ask for user details on the gateway form
PERIMETER_REQUIRE_USER_DETAILS = get_setting(
    "PERIMETER_REQUIRE_USER_DETAILS", False, cast_func=CAST_AS_BOOL
)



from __future__ import annotations

from typing import TYPE_CHECKING

from django import forms
from django.core.exceptions import ValidationError
from django.http import HttpRequest

from .models import AccessToken
from .settings import PERIMETER_SESSION_KEY

if TYPE_CHECKING:
    from .models import AccessTokenUse


class TokenGatewayForm(forms.Form):
    """Form used to process a perimeter request."""

    token = forms.CharField(required=True, max_length=100)

    def clean_token(self) -> AccessToken:
        """Validate the token against existing tokens."""
        try:
            _token = AccessToken.objects.get(token=self.cleaned_data.get("token"))
            if _token.has_expired:
                raise ValidationError("Token has expired", code="expired")
            if not _token.is_active:
                raise ValidationError("Token is inactive", code="invalid")
        except AccessToken.DoesNotExist:
            raise ValidationError("Token not found", code="invalid")
        else:
            self._token = _token
            return _token

    def save_token(self, request: HttpRequest) -> AccessTokenUse:
        """Record use of the token."""
        request.session[PERIMETER_SESSION_KEY] = self._token.token
        return self._token.record(
            user_email=self.cleaned_data.get("email"),
            user_name=self.cleaned_data.get("name"),
            client_ip=request.META.get("REMOTE_ADDR", "unknown"),
            client_user_agent=request.META.get("HTTP_USER_AGENT", "unknown"),
        )

    def save(self, request: HttpRequest) -> AccessTokenUse:
        """Create a new AccessTokenUse object from the form."""
        if getattr(self, "_token", None) is None:
            raise ValueError("Form token attr is not set")
        return self.save_token(request)


class UserGatewayForm(TokenGatewayForm):
    """Form used to process a perimeter request with user info."""

    email = forms.EmailField(required=True)
    name = forms.CharField(required=True, max_length=100)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['perimeter',
 'perimeter.management',
 'perimeter.management.commands',
 'perimeter.migrations']

package_data = \
{'': ['*'], 'perimeter': ['templates/perimeter/*']}

install_requires = \
['django>=2.2,<4.0']

setup_kwargs = {
    'name': 'perimeter',
    'version': '0.14',
    'description': 'Site-wide perimeter access control for Django projects.',
    'long_description': ".. image:: https://badge.fury.io/py/django-perimeter.svg\n    :target: https://badge.fury.io/py/django-perimeter\n\n.. image:: https://travis-ci.org/yunojuno/django-perimeter.svg?branch=master\n    :target: https://travis-ci.org/yunojuno/django-perimeter\n\n**This package now requires Python 3.7+ and Django 2.2+. For previous versions please refer to the relevant branch.**\n\nDjango Perimeter\n================\n\nPerimeter is a Django app that provides middleware that allows you to 'secure the perimeter' of your django site outside of any existing auth process that you have.\n\nWhy?\n----\n\nMost django sites have some kind of user registration and security model - a login process, decorators to secure certain URLs, user accounts - everything that comes with django.contrib.auth and associated apps (django-registration).\n\nSometimes, however, you want to simply secure the entire site to prevent prying eyes - the classic example being before a site goes live. You want to erect a secure perimeter fence around the entire thing. If you have control over your front-end web server (e.g. Apache, Nginx) then this can be used to do this using their in-built access control features. However, if you are running your app on a hosting platform you may not have admin access to these parts. Even if you do have control over your webserver, you may not want to be re-configuring it every time you want to grant someone access.\n\nThat's when you need Perimeter.\n\nPerimeter provides simple tokenised access control over your entire Django site (everything, including the admin site and login pages).\n\nHow does it work?\n-----------------\n\nOnce you have installed and enabled Perimeter, everyone requiring access will need an authorisation token (not authentication - there is nothing inherent in Perimeter to prevent people swapping / sharing tokens - that is an accepted use case).\n\nPerimeter runs as middleware that will inspect the user's ``session`` for a\ntoken. If they have a valid token, then they continue to use the site uninterrupted. If they do not have a token, or the token is invalid (expired or set to inactive), then they are redirected to the Perimeter 'Gateway', where they must enter a valid token, along with their name and email (for auditing purposes - this is stored in the database).\n\nTo create a new token you need to head to the admin site, and create a new token under the Perimeter app. If you have ``PERIMETER_ENABLED`` set to True already you won't be able to access the admin site (as Perimeter covers everything except for the perimeter 'gateway' form), and so there is a management command (``create_access_token``) that you can use to create your first token. (This is analagous to the Django setup process where it prompts you to create a superuser.)\n\nSetup\n-----\n\n1. Add 'perimeter' to your installed apps.\n2. Add 'perimeter.middleware.PerimeterAccessMiddleware' to the list of MIDDLEWARE_CLASSES\n3. Add the perimeter urls - NB must use the 'perimeter' namespace\n4. Add PERIMETER_ENABLED=True to your settings file. This setting can be used to enable or disable Perimeter in different environments.\n\n\nSettings:\n\n.. code:: python\n\n    PERIMETER_ENABLED = True\n\n    INSTALLED_APPS = (\n        ...\n        'perimeter',\n        ...\n    )\n\n    # perimeter must appear after sessions middleware as it relies on there\n    # being a valid request.session\n    MIDDLEWARE_CLASSES = [\n        ...\n        'django.contrib.sessions.middleware.SessionMiddleware',\n        'perimeter.middleware.PerimeterAccessMiddleware',\n        ...\n    ]\n\nSite urls:\n\n.. code:: python\n\n    # in site urls\n    urlpatterns = patterns(\n        '',\n        ...\n        # NB you must include the namespace, as it is referenced in the app\n        url(r'^perimeter/', include('perimeter.urls', namespace='perimeter')),\n        ...\n    )\n\nTests\n-----\n\nThe app has a suite of tests, and a ``tox.ini`` file configured to run them when using ``tox`` (recommended).\n",
    'author': 'YunoJuno',
    'author_email': 'code@yunojuno.com',
    'maintainer': 'YunoJuno',
    'maintainer_email': 'code@yunojuno.com',
    'url': 'https://github.com/yunojuno/django-perimeter',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
